/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package reflexion;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Parameter;
import java.lang.reflect.Type;
import java.time.LocalDateTime;
import java.util.ArrayList;

/**
 *
 * @author Krupi
 */
public class Inspector {
    /**
     * Array of Classes, that were used somewhere in the code, and have to be checked for need of import
     */
    private ArrayList<Class> imports;
    /**
     * An Object, that is used for inspection
     */
    private Object object;
    /**
     * A Class, that is used for inspection, or is the class of "object"
     */
    private Class classInstance;
    /**
     * Nubmer of exceptions, that appeared during the process of inspection
     */
    private int exceptions;
    
    /**
     * Creates a new Inspector, using an instance
     * @param object An instance of a Class, to which would be applied the inspection
     */
    public Inspector(Object object){
        this.object = object;
        this.classInstance = object.getClass();
        imports = new ArrayList<>();
        exceptions = 0;
    }
    /**
     * Creates a new Inspector, using a Class
     * @param c A Class, to which would be applied the inspection
     */
    public Inspector(Class c){
        this.classInstance = c;
        imports = new ArrayList<>();
        exceptions = 0;
    }
    /**
     * Collects data about the Class of the given instance, and returns them as a String
     * @return String containing collected data
     */
    private String getClassInfoAsString(){
        String text = "";
        //creates package info
        text += "package " + classInstance.getPackage().getName() + ";\n\n";
        //saves info about class to memory
        String classBody = getClassBody(classInstance, 0);
        boolean usedImportLayer = false;
        //controls all of imports, and creates info about them
        imports.sort((a, b) -> a.getName().compareTo(b.getName()));
        for(Class c : imports){
            boolean classMatch = false;
            for(Class cp : classInstance.getClasses()){
                if(c == cp){
                    classMatch = true;
                }
            }
            try{
                if(!classMatch){
                    if(!c.getPackage().equals(classInstance.getPackage()) && !c.getPackage().getName().equals("java.lang")){
                        text += "import " + c.getName() + ";\n";
                        usedImportLayer = true;
                    }
                }
            }
            catch(Exception e){
                exceptions ++;
            }
        }
        if(usedImportLayer)
            text += "\n";
        //loads class info
        text += classBody;
        return text;
    }
    /**
     * Collects data about the "body" (everything below line "public/private class/interface/enum xxx....." (including) adn returns them as a String
     * @param objectC Class to inspect
     * @param baseMargin Margin from the left side
     * @return String containing collected data
     */
    private String getClassBody(Class objectC, int baseMargin){
        String text = "";
        //creates margin
        text += addMargin(baseMargin);
        //creates class header
        text += createClassHeader(objectC) + "\n\n";
        //creates and sorts fields
        Field[] fields = objectC.getDeclaredFields();
        boolean usedStaticFields = false;
        for(Field f : fields){
            try{
                if(Modifier.isStatic(f.getModifiers())){
                    text += createField(f, baseMargin + 4, objectC);
                    usedStaticFields = true;
                }
            }
            catch(Exception e){
                exceptions ++;
            }
        }
        if(usedStaticFields)
            text += "\n";
        boolean usedNonStaticFields = false;
        for(Field f : fields){
            try{
                if(!Modifier.isStatic(f.getModifiers())){
                    text += createField(f, baseMargin + 4, objectC);
                    usedNonStaticFields = true;
                }
            }
            catch(Exception e){
                exceptions ++;
            }
        }
        if(usedNonStaticFields)
            text += "\n";
        //creates contructors
        Constructor[] constructors = objectC.getDeclaredConstructors();
        for(Constructor c : constructors){
            try{
                text += createConstructor(c, baseMargin + 4, objectC);
            }
            catch(Exception e){
                exceptions ++;
            }
        }
        text += "\n";
        //creates and sorts methods
        Method[] methods = objectC.getDeclaredMethods();
        boolean usedStaticMethods = false;
        for(Method m : methods){
            try{
                if(Modifier.isStatic(m.getModifiers())){
                    text += createMethod(m, baseMargin + 4, objectC);
                    usedStaticMethods = true;
                }
            }
            catch(Exception e){
                exceptions ++;
            }
        }
        if(usedStaticMethods)
            text += "\n";
        boolean usedNonStaticMethods = false;
        for(Method m : methods){
            try{
                if(!Modifier.isStatic(m.getModifiers())){
                    text += createMethod(m, baseMargin + 4, objectC);
                    usedNonStaticMethods = true;
                }
            }
            catch(Exception e){
                exceptions ++;
            }
        }
        if(usedNonStaticMethods)
            text += "\n";
        for(Class c : objectC.getDeclaredClasses()){
            text += getClassBody(c, baseMargin + 4) + "\n";
        }
        text += addMargin(baseMargin);
        text += "}\n";
        return text;
    }
    /**
     * Collects data about Class of given instance, and output them
     * @param type Target of the output, if null, only returns String for other purposes
     * @return String, containting collected data
     */
    public String getClassInfo(OutputType type){
        //resets counter of exceptions
        exceptions = 0;
        //calls main inspecting method
        String text = getClassInfoAsString();
        //does nothing, only return String
        if(type == null){
            return text;
        }
        //prints output into Console
        else if(type == OutputType.Console){
            System.out.println(text);
            //prints info about exceptions
            if(exceptions != 0){
                System.out.println("========================================================================================\nThere were " + exceptions + " problems while inspecting class, some of the data may be missing.\n========================================================================================");
            }
        }
        //saves output on disk in file in APPDATA, indentified as NAME_OF_CLASS-DATE-TIME
        else if(type == OutputType.File){
            String path = System.getenv("APPDATA") + File.separator + "Inspector_Output";
            File file = new File(path);
            if(!file.exists()){
                if(!file.mkdir()){
                    System.err.println("Creation of a file \"" + path + "\", was unsuccessful, check the sucurity settings.");
                    return text;
                }
            }
            LocalDateTime time = LocalDateTime.now();
            String outputPath = path + File.separator + classInstance.getSimpleName() + "_" + time.getDayOfMonth() + "-" + time.getMonthValue() + "-" + time.getYear() + "_" + time.getHour() + "-" + time.getMinute() + "-" + time.getSecond() + "-" + time.getNano() + ".txt";
            try(BufferedWriter out = new BufferedWriter(new FileWriter(outputPath))){
                String[] lines = text.split("\n");
                for(String s : lines){
                    out.write(s);
                    out.newLine();
                    out.flush();
                }
                System.out.println("File saved at: " + outputPath);
                if(exceptions != 0){
                    System.out.println("========================================================================================\nThere were " + exceptions + " problems while inspecting class, some of the data may be missing.\n========================================================================================");
                }
            } catch (IOException ex) {
                System.err.println("Creation of a file \"" + outputPath + "\", was unsuccessful, check the sucurity settings.");
            }
        }
        return text;
    }
    /**
     * Creates the "header" line of a Class source code.
     * @param c Inpected Class
     * @return String, containing header (firts line)
     */
    private String createClassHeader(Class c){
        String header = "";
        header += c.toGenericString().replace(c.getPackage().getName() + ".", "") + " ";
        //repairs genericity modified Class name in "private" classes
        if(classInstance != c){
            String name = classInstance.getSimpleName();
            header = header.replace(name + "$", "");
        }
        //creates info about superclass
        Class superclass = c.getSuperclass();
        if (!superclass.getSimpleName().equals("Object")){
            addImport(superclass);
            header += "extends ";
            String superString = convertGeneric(superclass.toGenericString());
            if(classInstance != c){
                String name = classInstance.getSimpleName();
                superString = superString.replace(name + "$", "");
            }
            header += superString + " ";
        }
        //creates info about interfaces
        if(c.getInterfaces().length != 0){
            header += "implements ";
            for (Type t : c.getGenericInterfaces()){
                header += convertGeneric(t.getTypeName());
                if (!t.equals(c.getGenericInterfaces()[c.getGenericInterfaces().length - 1])){
                    header += ", ";
                }
                else {
                    header += " ";
                }
                
            }
            for(Class i : c.getInterfaces()){
                addImport(i);
            }
        }
        header += "{";
        return header;
    }
    /** 
     * Add Class to the queue for controlling packages and writing imports to the output String
     * @param c Class to find import to
     */
    private void addImport(Class c){
        boolean b = true;
        //replaces "collection" classes with normal
        if(c.getName().startsWith("[L")){
            String newName = c.getClass().getName().replace("[L", "");
            try {
                c = Class.forName(newName);
            } catch (ClassNotFoundException ex) {
                
            }
        }
        //checks, if the object isn't already on the list, and adds it there
        if(!c.getSimpleName().equals("Object") && c.getName().contains(".")){
            for(Class i : imports){
                if (i == c)
                    b = false;
            }
            if(b)
                imports.add(c);
        }
    }
    /**
     * Collects data about the given field, and returns them as a String
     * @param field A field to inspect
     * @param margin Margin from the left side
     * @param c Class, that is the field taken from
     * @return String containing collected data
     */
    private String createField(Field field, int margin, Class c){
        String text = "";
        text += addMargin(margin);
        String fieldS = field.toGenericString().replace(c.getCanonicalName() + ".", "");
        if(!field.getType().isPrimitive()){
            //removes package prefix from class name
            if(field.getType().getCanonicalName() != null)
                fieldS = fieldS.replace(field.getType().getCanonicalName(), field.getType().getSimpleName());
        }
        //handle problems of names of "private" classes
        if(classInstance != c && (fieldS.contains("java")))
            fieldS = fieldS.replace(fieldS.substring(fieldS.indexOf("java"), fieldS.lastIndexOf(".") + 1), "");
        else if(classInstance != c){
            fieldS = fieldS.replace(fieldS.substring(fieldS.indexOf(classInstance.getSimpleName()), fieldS.lastIndexOf(".") + 1), "");
        }
        text += fieldS;
        if(!field.getType().isPrimitive()){
            addImport(field.getType());
        }
        try {
            field.setAccessible(true);
            String value = field.get(object).toString().replace("\n", "\\n").replace("\r", "\\r");
            text += " = " + value;
        } catch (Exception ex) {
            
        }
        text += ";\n";
        return text;
    }
    /**
     * Collects data about the given method, and returns them as a String
     * @param method A method to inspect
     * @param margin Margin from the left side
     * @param c Class, that is the method taken from
     * @return String containing collected data
     */
    private String createMethod(Method method, int margin, Class c){
        String text = "";
        text += addMargin(margin);
        String methodS = method.toGenericString().replace(c.getCanonicalName() + ".", "");
        if(!method.getReturnType().isPrimitive()){
            if(method.getReturnType().getCanonicalName() != null)
                methodS = methodS.replace(method.getReturnType().getCanonicalName(), method.getReturnType().getSimpleName());
            methodS = methodS.replace(method.getReturnType().getName(), method.getReturnType().getSimpleName());
            addImport(method.getReturnType());
        }
        for(Parameter p : method.getParameters()){
            if(!p.getType().isPrimitive()){
                addImport(p.getType());
            }
            if(p.getType().getCanonicalName() != null){
                methodS = methodS.replace(p.getType().getCanonicalName(), p.getType().getSimpleName());
            }
        }
        for(Class e : method.getExceptionTypes()){
            if(e.getCanonicalName() != null)
                methodS = methodS.replace(e.getCanonicalName(), e.getSimpleName());
        }
        if(classInstance != c){
            methodS = methodS.replace(methodS.substring(methodS.lastIndexOf("java"), methodS.lastIndexOf(".") + 1), "");
        }
        text += methodS + "{}\n";
        return text;
    }
    /**
     * Collects data about the given constructor, and returns them as a String
     * @param constructor A constructor to inspect
     * @param margin Margin from the left side
     * @param c Class, that is the constructor taken from
     * @return String containing collected data
     */
    private String createConstructor(Constructor constructor, int margin, Class c){
        String text = "";
        text += addMargin(margin);
        String constructorS = constructor.toGenericString().replace(c.getCanonicalName(), c.getSimpleName());
        for(Parameter p : constructor.getParameters()){
            if(!p.getType().isPrimitive()){
                if(p.getType().getCanonicalName() != null)
                    constructorS = constructorS.replace(p.getType().getCanonicalName(), p.getType().getSimpleName());
                addImport(p.getType());
            }
        }
        for(Class e : constructor.getExceptionTypes()){
            if(e.getCanonicalName() != null)
                constructorS = constructorS.replace(e.getCanonicalName(), e.getSimpleName());
        }
        if(classInstance != c){
            constructorS = constructorS.replaceFirst(constructorS.substring(constructorS.indexOf(classInstance.getSimpleName()), constructorS.indexOf(c.getSimpleName())), "");
            constructorS = constructorS.replace(constructorS.substring(0, constructorS.indexOf("(")), c.getSimpleName());
        }
        text += constructorS;
        text += "{}\n";
        return text;
    }
    /**
     * Creates a String, with given space, used in margin
     * @param margin Amount of space to enter in the String
     * @return String, containing give amount of space
     */
    private String addMargin(int margin){
        String text = "";
        for(int i = 0; i < margin; i++){
            text += " ";
        }
        return text;
    }
    /**
     * Removes package prefix from generic Class String (used in Class header)
     * @param s String to cut
     * @return String, containing Class Name with genericity definition 
     */
    private String convertGeneric(String s){
        return s.replace(s.substring(0, s.lastIndexOf(".") + 1), "");
    }
}
